import imspy_connector
ims = imspy_connector.py_constants

MASS_PROTON = ims.mass_proton()
MASS_NEUTRON = ims.mass_neutron()
MASS_ELECTRON = ims.mass_electron()
MASS_WATER = ims.mass_water()
STANDARD_TEMPERATURE = ims.standard_temperature()
STANDARD_PRESSURE = ims.standard_pressure()
ELEMENTARY_CHARGE = ims.elementary_charge()
K_BOLTZMANN = ims.k_boltzmann()
AVOGADRO_NUMBER = ims.avogadro()
